#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['fastapi_users',
 'fastapi_users.authentication',
 'fastapi_users.authentication.strategy',
 'fastapi_users.authentication.strategy.db',
 'fastapi_users.authentication.transport',
 'fastapi_users.db',
 'fastapi_users.router']

package_data = \
{'': ['*']}

install_requires = \
['fastapi >=0.65.2',
 'passlib[bcrypt] ==1.7.4',
 'email-validator >=1.1.0,<1.3',
 'pyjwt[crypto] ==2.4.0',
 'python-multipart ==0.0.5',
 'makefun >=1.11.2,<2.0.0']

extras_require = \
{":python_version < '3.8'": ['typing-extensions >=4.1.1'],
 'beanie': ['fastapi-users-db-beanie >=1.0.0'],
 'dev': ['flake8',
         'pytest',
         'requests',
         'isort',
         'pytest-asyncio',
         'flake8-docstrings',
         'mike',
         'mkdocs',
         'mkdocs-material',
         'mkdocs-mermaid2-plugin',
         'black',
         'mypy',
         'codecov',
         'pytest-cov',
         'pytest-mock',
         'flit',
         'markdown-include',
         'pygments',
         'pymdown-extensions',
         'bumpversion',
         'httpx-oauth',
         'httpx',
         'asgi_lifespan',
         'uvicorn',
         'types-redis'],
 'oauth': ['httpx-oauth >=0.4,<0.8'],
 'redis': ['redis >=4.3.3,<5.0.0'],
 'sqlalchemy': ['fastapi-users-db-sqlalchemy >=4.0.0']}

setup(name='fastapi-users',
      version='10.1.5',
      description='Ready-to-use and customizable users management for FastAPI',
      author=None,
      author_email='François Voron <fvoron@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )
